/*==============================================================================


   Purpose:    hardware performance monitor command as on CRAY vector machines
               based on the PCL Performance Counter Library

   Author:     Rudolf Berrendorf
               University of Applied Sciences Bonn-Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-bonn-rhein-sieg.de


   Date:       04.07.2000



   Copyright (C) 1999, 2000
   Forschungszentrum Juelich GmbH, Federal Republic of Germany.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.


==============================================================================*/
/*============================================================================*/
/* hpm takes as argument an executable program (including arguments) and      */
/* measures performance relevant events. People ever used the hpm on CRAY     */
/* vector machines know what I mean.                                          */
/* Example:                                                                   */
/*   hpm -e PCL_FP_INSTR a.out 500 200                                        */
/* start the program a.out with arguments 500 and 200 with performance        */
/* monitoring the floating point instructions executed in a.out.              */
/*                                                                            */
/*============================================================================*/
/* include files                                                              */


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>

#include "pcl.h"
#include "processor_number.h"


#if defined(ALPHA_DEC) || defined(MIPS) || defined(POWERPC) || defined(SR8000)
static char *
strsignal(int signo)
{
  static char buf[80];
  sprintf(buf, "signal %d raised", signo);
  return buf;
}

#elif defined(PENTIUM_LINUX)
/* for some reason, it's not declared in string.h */
char *strsignal(int sig);

#endif


/*                                                                            */
/*============================================================================*/
/* environment variables                                                      */


/* environment variable containg the events to measure.
   The format is event_1,...,event_n. No blanks etc.
   Example for ksh: export HPM_EVENTS=PCL_CYCLES,PCL_FP_INSTR
   Command line options -e and --events overwrite this variable.
 */
#define HPM_EVENTS "HPM_EVENTS"

/* environment variable containg the mode in which events should be counted.
   Example for ksh: export HPM_COUNTMODE=PCL_MODE_USER
   Command line options -m and --mode overwrite this variable.
 */
#define HPM_COUNTMODE "HPM_COUNTMODE"


/*                                                                            */
/*============================================================================*/
/* global variables and constants                                             */


/* hpm version number string */
static const char *version_string = "hpm v1.0";

/* contains the program name for the supervisor program. */
static char *program_name;

/* set to 1 when child process stopped */
static volatile int child_stopped = 0;

/* be verbose */
static int verbose;

/* start and stop time for the executed program */
static struct timeval start_time, stop_time;

/* string buffer for event names to measure */
#define MAX_EVENT_NAME_SIZE 8192
static char event_names[MAX_EVENT_NAME_SIZE];

/* string buffer for count mode */
#define MAX_COUNTMODE_SIZE 8192
static char count_mode_name[MAX_COUNTMODE_SIZE];

/* name of output file */
static char outfilename[FILENAME_MAX];
static FILE *outfile;


/* variables for calling PCL library */
/* events to measure */
static int events[PCL_COUNTER_MAX];

/* number of events to measure */
static int nevents;

/* count mode */
static unsigned int mode;

/* handle */
static void *descr;


/*                                                                            */
/*============================================================================*/
/* print error message and stop if status == 0                                */


static void
error(int status, int errnum, char *message)
{
  fflush(outfile);
  fprintf(outfile, "%s: %s (%d)\n", program_name, message, errnum);
  fflush(outfile);
  if (status != 0)
    exit(status);
}


/*                                                                            */
/*============================================================================*/
/* returns the PCL constant for an event given as a string                    */
/* i.e. returns PCL_CYCLES for the string "PCL_CYCLES"                        */


static int
map_eventnames(char *name)
{
  int i;

  /* search all names */
  for (i = 0; i < PCL_MAX_EVENT; ++i)
    if (!strcmp(name, PCLeventname(i)))
      return i;

  /* not found */
  return -1;
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */


static int
parse_events(char *event_names, int events[], int *nevents)
{
  char buf[MAX_EVENT_NAME_SIZE], *bufptr;
  int event;


  /* handle events */
  *nevents = -1;

  /* loop over string */
  while (*event_names)
    {
      bufptr = buf;

      /* loop over one name () */
      while (*event_names && (*event_names != ','))
	*bufptr++ = *event_names++;
      *bufptr = '\0';

      /* look for event */
      if ((event = map_eventnames(buf)) == -1)
	return -1;

      /* one more event */
      ++*nevents;
      events[*nevents] = event;

      /* skip , */
      if (*event_names == ',')
	++event_names;
    }

  ++*nevents;


  /* handle count mode */
  if (!strcmp(count_mode_name, "PCL_MODE_USER"))
    mode = PCL_MODE_USER;
  else if (!strcmp(count_mode_name, "PCL_MODE_SYSTEM"))
    mode = PCL_MODE_SYSTEM;
  else if (!strcmp(count_mode_name, "PCL_MODE_USER_SYSTEM"))
    mode = PCL_MODE_USER_SYSTEM;
  else
    fprintf(outfile, "%s: illegal count mode\n", program_name);


  /* open output file */
  if (!strcmp(outfilename, ""))
    outfile = stdout;
  else
    {
      if ((outfile = fopen(outfilename, "w")) == NULL)
	return -1;
    }

  return 0;
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */


static int
start_counting(void)
{
  int i, errnum;


  /* get start time */
  gettimeofday(&start_time, (struct timezone *) 0);

  /* start PCL */
  if (verbose)
    {
      fprintf(outfile, "%s: starting PCL with mode=%d and events=",
	      program_name, mode);
      for (i = 0; i < nevents; ++i)
	fprintf(outfile, "%d ", events[i]);
      fprintf(outfile, "\n");
    }
  if ((errnum = PCLstart(descr, events, nevents, mode)) != PCL_SUCCESS)
    error(1, errnum, "unable to start PCL");

  return 0;
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */


static int
stop_counting(void)
{
  int errnum, i;
  PCL_CNT_TYPE i_result[PCL_COUNTER_MAX];
  PCL_FP_CNT_TYPE fp_result[PCL_COUNTER_MAX];


  /* stop PCL */
  if ((errnum = PCLstop(descr, i_result, fp_result, nevents)) != PCL_SUCCESS)
    error(1, errnum, "error in stopping PCL");

  /* get end time */
  gettimeofday(&stop_time, (struct timezone *) 0);


  /* print results */
  fprintf(outfile,
	  "%s: elapsed time: %.3f s\n",
	  program_name,
	  (double) ((stop_time.tv_sec - start_time.tv_sec)
		    + (stop_time.tv_usec - start_time.tv_usec) * 0.000001));

  for (i = 0; i < nevents; ++i)
    {
      if (events[i] >= PCL_MFLOPS)
	{
	  /* results are 64-bit floating point values */
	  fprintf(outfile,
		  "%s: counter %d   : %f %s\n",
		  program_name, i, fp_result[i], PCLeventname(events[i]));
	}
      else
	{
	  fprintf(outfile,
		  "%s: counter %d   : %.0f %s\n",
		  program_name, i, (double) i_result[i],
		  PCLeventname(events[i]));
	}
    }

  return 0;
}


/*                                                                            */
/*============================================================================*/
/* catches signal that child process terminates                               */


static void
child_watcher(int signo)
{
  /* set binary variable to 1 */
  child_stopped = 1;
}


/*                                                                            */
/*============================================================================*/
/* run a command under hpm control                                            */


static void
run_command(char **cmd)
{
  typedef void (*sighandler) (int);
  sighandler interrupt_signal, quit_signal;
  pid_t pid;
  int status, processor_number;
  char msg[81];
  PCL_CNT_TYPE i_result[1];
  PCL_FP_CNT_TYPE fp_result[1];
  int event, res;
  unsigned int mode;


  /* trick: call PCLstart/stop to bind process to processor on some systems */
  mode = PCL_MODE_USER;
  event = 0;
  while ((event < PCL_MAX_EVENT)
	 && (PCLquery(descr, &event, 1, mode) != PCL_SUCCESS))
    ++event;
  if (event == PCL_MAX_EVENT)
    error(1, 0, "unable to start PCL");
  if (((res = PCLstart(descr, &event, 1, mode) != PCL_SUCCESS))
      || ((res = PCLstop(descr, i_result, fp_result, 1) != PCL_SUCCESS)))
    error(1, res, "unable to start PCL");

  /* get processor number we are currently running on */
  processor_number = PCL_processor_number();


  /* establish signal handler to watch for child's death */
  if (signal(SIGCHLD, child_watcher) == SIG_ERR)
    error(0, errno, "signal() failed");

  /* start performance counting */
  if ((status = start_counting()) != 0)
    error(1, status, "unable to start PCL");


  if (verbose)
    fprintf(outfile, "%s: forking child process on processor %d\n",
	    program_name, processor_number);
  pid = fork();
  if (pid < 0)
    /* error happened */
    error(1, errno, "cannot fork");
  else if (pid == 0)
    {
      if (verbose)
	fprintf(outfile, "%s: child started on processor %d\n",
		program_name, processor_number);

      /* check that we are running on the same processor */
      if (PCL_processor_number() == processor_number)
	/* child process: execute the user command */
	execvp(cmd[0], cmd);
      else
	printf("child started on different processor (%d) than father (%d)\n",
	       PCL_processor_number(), processor_number);

      /* this line is reached only when the last line couldn't be executed
         without error */
      sprintf(msg, "cannot run %-69s", cmd[0]);
      error(0, errno, msg);
      _exit(errno == ENOENT ? 127 : 126);
    }


  /* Have signals kill the child but not self (if possible).  */
  interrupt_signal = signal(SIGINT, SIG_IGN);
  quit_signal = signal(SIGQUIT, SIG_IGN);

  /* wait for child to terminate */
  if (verbose)
    fprintf(outfile, "%s: waiting for pid=%d\n", program_name, (int) pid);
  while (pause() && !child_stopped)
    /* keep process running as in PCL other signals might be used */
    ;
  if (verbose)
    fprintf(outfile, "%s: child terminated\n", program_name);

  /* wait for the child process even if it has already stopped as we need the
     exit status of the child */
  if (waitpid(pid, &status, 0) != pid)
    {
      fprintf(outfile, "problem with waitpid\n");
    }
  else
    {
      if (WIFSIGNALED(status) != 0)
	{
	  /* child process exited because of uncaught signal */
	  fprintf(outfile, "%s\n", strsignal(WTERMSIG(status)));
	}
    }
  if (verbose)
    fprintf(outfile, "%s: waitpid for child succeeded. Exit status=%d\n",
	    program_name, WIFEXITED(status));


  /* stop counting */
  if ((status = stop_counting()) != 0)
    error(1, status, "error waiting for child process");

  /* Re-enable signals.  */
  signal(SIGINT, interrupt_signal);
  signal(SIGQUIT, quit_signal);
}


/*                                                                            */
/*============================================================================*/
/* print usage information                                                    */


static void
usage(FILE * stream, int status)
{
  fprintf(stream,
	  "Usage: %s [hpm-options] command [cmd-args]\n"
	  "where hpm-options might be  ( (o1|o2) means o1 or o2):\n"
	  "\t(-e | --events) event_1,...,event_n : define events to measure\n"
	  "\t(-m | --mode) count_mode            : define count-mode\n"
	  "\t(-o | --output) filename            : direct output to file\n"
	  "\t(-h | --help)                       : show this\n"
	  "\t(-s | --show-events)                : show possible events on "
	  " this machine\n"
	  "\t(-v | --verbose)                    : be verbose\n"
	  "\t(-V | --version)                    : show program version\n"
	  "where 'event' may be any combination of PCL-events (see man pcl)\n",
	  program_name);
  exit(status);
}


/*                                                                            */
/*============================================================================*/
/* parse command line arguments and preset global variables                   */

static char **
getargs(int argc, char **argv)
{
  int optc, event;
  char *env;


  /* preset global variables */
  verbose = 0;
  program_name = argv[0];
#if defined(MIPS)
  strcpy(count_mode_name, "PCL_MODE_USER");
#else
  strcpy(count_mode_name, "PCL_MODE_USER_SYSTEM");
#endif
  strcpy(event_names, "PCL_CYCLES");
  strcpy(outfilename, "");
  outfile = NULL;


  /* check for environment variable for events to measure (might be overwritten
     below) */
  if ((env = getenv(HPM_EVENTS)) != NULL)
    {
      strncpy(event_names, env, MAX_EVENT_NAME_SIZE);
      if (verbose)
	fprintf(outfile, "%s: reading environment variable %s=%s\n",
		program_name, HPM_EVENTS, env);
    }

  /* check for environment variable for count mode (might be overwritten below) 
   */
  if ((env = getenv(HPM_COUNTMODE)) != NULL)
    {
      strncpy(count_mode_name, env, MAX_COUNTMODE_SIZE);
      if (verbose)
	fprintf(outfile, "%s: reading environment variable %s=%s\n",
		program_name, HPM_COUNTMODE, env);
    }

  /* parse command line arguments until we find an unknown string on command
     line */
  for (optc = 1; optc < argc; ++optc)
    {
      /* specify events */
      if ((!strcmp(argv[optc], "-e")) || (!strcmp(argv[optc], "--events")))
	{
	  /* next argument gives events to measure */
	  ++optc;
	  if (optc == argc)
	    usage(stdout, 1);
	  strncpy(event_names, argv[optc], MAX_EVENT_NAME_SIZE);
	}

      /* specify count mode */
      else if ((!strcmp(argv[optc], "-m")) || (!strcmp(argv[optc], "--mode")))
	{
	  /* next argument gives count mode */
	  ++optc;
	  if (optc == argc)
	    usage(stdout, 1);
	  strncpy(count_mode_name, argv[optc], MAX_COUNTMODE_SIZE);
	}

      /* specify output file */
      else if ((!strcmp(argv[optc], "-o")) || (!strcmp(argv[optc], "--output")))
	{
	  /* next argument gives output filename */
	  ++optc;
	  if (optc == argc)
	    usage(stdout, 1);
	  strncpy(outfilename, argv[optc], FILENAME_MAX);
	}

      /* give help */
      else if ((!strcmp(argv[optc], "-h")) || (!strcmp(argv[optc], "--help")))
	usage(stdout, 0);

      /* give help */
      else if ((!strcmp(argv[optc], "-s"))
	       || (!strcmp(argv[optc], "--show-events")))
	{
	  printf("%s: possible events on this machine:\n", program_name);
	  for (event = 0; event < PCL_MAX_EVENT; ++event)
	    if (PCLquery(descr, &event, 1, PCL_MODE_USER) == PCL_SUCCESS)
	      printf("\t%-25s\n", PCLeventname(event));
	  exit(1);
	}

      /* be verbose */
      else if ((!strcmp(argv[optc], "-v"))
	       || (!strcmp(argv[optc], "--verbose")))
	verbose = 1;

      /* show version */
      else if ((!strcmp(argv[optc], "-V"))
	       || (!strcmp(argv[optc], "--version")))
	{
	  fprintf(outfile, "%s\n", version_string);
	  exit(0);
	}

      else
	{
	  if (argv[optc][0] == '-')
	    /* some unknown option */
	    usage(stderr, 1);
	  break;
	}
    }

  /* At this point we have found a string in the command line which is not a
     known option for us or the rest of the command line is empty. */
  if (optc == argc)
    usage(stderr, 1);

  /* parse events names and build structure for PCLstart */
  if (parse_events(event_names, events, &nevents) == -1)
    usage(stdout, 1);

  /* return the rest of the command line */
  return (char **) &argv[optc];
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */


int
main(int argc, char **argv)
{
  /* command line for the program */
  char **command_line;
  int errnum;


  if ((errnum = PCLinit(&descr)) != PCL_SUCCESS)
    error(1, errnum, "unable to allocate PCL descriptor");

  /* first set output file (may be overwritten later) */
  outfile = stdout;

  /* split command line arguments in those for hpm an those for the program */
  command_line = getargs(argc, argv);

  /* run the command */
  run_command(command_line);

  if ((errnum = PCLexit(descr)) != PCL_SUCCESS)
    error(1, errnum, "unable to deallocate PCL descriptor");

  return 0;
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
