/*==============================================================================


   Purpose:    low level routines for SPARC


   Author:     Rudolf Berrendorf
               University of Applied Sciences Bonn-Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-bonn-rhein-sieg.de


   last mod:   25.09.2001


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 2000
   Forschungszentrum Juelich GmbH, Federal Republic of Germany.
   Copyright (C) 2001
   University of Applied Sciences Bonn-Rhein-Sieg
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.
  
==============================================================================*/

#include "../../defines.h"
#include "../../pcl.h"
#include "hw_driver_defs.h"
#include "../../pcl_hw_driver.h"
#include "../../hw_driver_global_defs.h"
#include "../../state_descr.h"
#include "../../determine_mhz.h"

#include <fcntl.h>
#include <unistd.h>
#include <stropts.h>
#include <sys/types.h>
#include <sys/processor.h>
#include <sys/procset.h>

/* we distinguish between two system interfaces:
   perfmon    : <= Solaris 7 (support through perfmon package)
   non-perfmon: >= Solaris 8 (system support for hw performance counters)
*/
#include "perfmon.h"
#if !defined(PERFMON)
#include <libcpc.h>
#endif


/*                                                                            */
/*----------------------------------------------------------------------------*/
/* global variables                                                           */

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1,-1},{OP_NO,OP_NO,OP_NO}}}

/* UltraSPARC1 and UltraSPARC2 */
static const PCL_event_descr_t event_table_ultrasparc1[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x9,-1,-1},{OP_ADD,OP_NO,OP_NO}}}},  /* PCL_L1DCACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{0xa,-1,-1},{OP_ADD,OP_NO,OP_NO}}}}, /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0xf,-1,-1},{OP_ADD,OP_NO,OP_NO}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1ICACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{0x8,-1,-1},{OP_ADD,OP_NO,OP_NO}}}},  /* PCL_L1ICACHE_READWRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x8,-1},{OP_NO,OP_ADD,OP_NO}}}}, /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,COMPOSED,{{{0x8,0x8,-1},{OP_ADD,OP_SUB,OP_NO}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{0xc,-1,-1},{OP_ADD,OP_NO,OP_NO}}}},  /* PCL_L2CACHE_READWRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,0xc,-1},{OP_NO,OP_ADD,OP_NO}}}},  /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,0xd,-1},{OP_NO,OP_ADD,OP_NO}}}}, /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ITLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x0,0x0,-1},{OP_ADD,OP_ADD,OP_NO}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x00},{OP_NO,OP_NO,OP_ADD}}}},        /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_INTEGER_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x0,-1,-1},{OP_ADD,OP_NO,OP_NO}}}},  /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_JUMP_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_JUMP_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL */
  {PCL_UNAVAIL, PCL_INVALID},                      /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{0x1,0x0,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,OP_NO}}}}, /* PCL_IPC */
  {PCL_UNAVAIL, PCL_INVALID},                      /* PCL_L1DCACHE_MISSRATE */
  {PCL_AVAIL,COMPOSED,{{{0xf,0xd,-1},{OP_FP_1|OP_FP_2|OP_FP_REZ,OP_FP_2|OP_FP_MUL,OP_NO}}}}, /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                       /* PCL_MEM_FP_RATIO */
};

/*----------------------------------------------------------------------------*/

/* UltraSPARC3 */
static const PCL_event_descr_t event_table_ultrasparc3[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{9,0,-1},{OP_ADD,OP_NO,OP_NO}}}}, /* PCL_L1DCACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{10,0,-1},{OP_ADD,OP_NO,OP_NO}}}}, /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                      /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                      /* PCL_L1DCACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{8,0,-1},{OP_ADD,OP_NO,OP_NO}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                      /* PCL_L1ICACHE_READWRITE */
  {PCL_AVAIL,COMPOSED,{{{8,8,-1},{OP_ADD,OP_SUB,OP_NO}}}}, /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,8,-1},{OP_NO,OP_ADD,OP_NO}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{12,-1,-1},{OP_ADD,OP_NO,OP_NO}}}}, /* PCL_L2CACHE_READWRITE */
  {PCL_AVAIL,COMPOSED,{{{12,12,-1},{OP_ADD,OP_SUB,OP_NO}}}}, /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,12,-1},{OP_NO,OP_ADD,OP_NO}}}}, /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,17,-1},{OP_NO,OP_ADD,OP_NO}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,18,-1},{OP_NO,OP_ADD,OP_NO}}}}, /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0,0,-1},{OP_ADD,OP_ADD,OP_NO}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0},{OP_NO,OP_NO,OP_ADD}}}}, /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,COMPOSED,{{{24,39,-1},{OP_ADD,OP_ADD,OP_NO}}}}, /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{1,1,-1},{OP_ADD,OP_ADD,OP_NO}}}}, /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                       /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,COMPOSED,{{{21,29,-1},{OP_ADD,OP_ADD,OP_NO}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{22,-1,-1},{OP_ADD,OP_NO,OP_NO}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_LOAD */
  {PCL_AVAIL,NOT_COMP,{{{5,-1,-1},{OP_ADD,OP_NO,OP_NO}}}}, /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{24,39,0},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_ADD,OP_FP_2|OP_FP_DIV_MHZ}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{1,0,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,OP_NO}}}}, /* PCL_IPC */
  {PCL_UNAVAIL, PCL_INVALID},                      /* PCL_L1DCACHE_MISSRATE */
  {PCL_AVAIL,COMPOSED,{{{12,12,-1},{OP_FP_1|OP_FP_2|OP_FP_REZ,OP_FP_2|OP_FP_MUL,OP_NO}}}}, /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                       /* PCL_MEM_FP_RATIO */
};


/*                                                                            */
/*============================================================================*/
/* initialization                                                             */

int
PCL_driver_open(
		PCL_DESCR_TYPE d /* I/O: address of descriptor */
		)

{
  processorid_t proc;
  PCL_state_descr *descr = d;
  HW_state_descr *hw_descr;
#if !defined(PERFMON)
  const char *cpu_str;
#endif


  /* get state descriptor */
  hw_descr = &descr->hw_state_descr;

  /* initialize state variables */
  descr->allowed_count_mode = PCL_MODE_USER_SYSTEM;
  descr->event_table =  event_table_ultrasparc1;
  hw_descr->mhz_rate = PCL_determine_mhz_rate();

  /* query on what processor we are running */
  if (processor_bind(P_LWPID, P_MYID, PBIND_QUERY, &proc) == -1)
    return PCL_FAILURE;

  /* if not assigned to a CPU, take first available CPU */
  if (proc == -1)
    for(proc = 0; p_online(proc, P_STATUS) != P_ONLINE; ++proc)
      ;

  /* bind process to processor */
  if (processor_bind(P_LWPID, P_MYID, proc, NULL) == -1)
    return PCL_FAILURE;


#if defined(PERFMON)
  /* Try to open a channel to the perfmon driver */
  hw_descr->fd = open("/dev/perfmon", O_RDONLY);
  if (hw_descr->fd == -1)
    return PCL_FAILURE;

#else  /* defined(PERFMON) */
  /* check if compiled cpc version is supported by cpc library version */
  if(cpc_version(CPC_VER_CURRENT) == CPC_VER_NONE)
    /* version mismatch: library cannot translate */
    return PCL_FAILURE;

  /* check what processor we are running on */
  if(((hw_descr->cpuver = cpc_getcpuver()) == -1)
     || ((cpu_str = cpc_getcciname(hw_descr->cpuver)) == NULL))
    return PCL_FAILURE;

  /* determine CPU type */
  if(!strcmp(cpu_str, "UltraSPARC I&II"))
    descr->event_table =  event_table_ultrasparc1;
  else if(!strcmp(cpu_str, "UltraSPARC III"))
    descr->event_table =  event_table_ultrasparc3;
  else
    return PCL_FAILURE;

  /* check if we have access to the performance counters */
  if(cpc_access() == -1)
    return PCL_FAILURE;

#endif  /* defined(PERFMON) */


  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Starts hardware counters.                                                  */


int
PCL_driver_start(
		 PCL_DESCR_TYPE d, /* I: descriptor */
		 int max_counter,
		 PCL_BIT_MASK_TYPE counter_used_mask,
		 PCL_DRIVER_COMMAND_TYPE *commands,
		 unsigned int count_mode
		 )

{
  int sel0, sel1;
  PCL_BIT_MASK_TYPE used_mask;
  PCL_state_descr *descr = d;
  HW_state_descr *hw_descr;
#if defined(PERFMON)
  unsigned long long set_count, val;
#else
  cpc_event_t cpc_cmd;
#endif


  /* get state descriptor */
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];


  /* initialize selectors with default count events (cycles, cycles) */
  sel0 = sel1 = 0;

#if defined(PERFMON)
  /* Extract selected events */
  if(IS_USED(used_mask, 0))
    sel0 = commands[0] << 4;
  if(IS_USED(used_mask, 1))
    sel1 = commands[1] << 11;

  /* Set PCR to accumulate desired events */
  set_count = sel0 | sel1;

  if(count_mode == PCL_MODE_USER)
    /* user mode only */
    set_count |= PCR_USER_TRACE;
  else if(count_mode == PCL_MODE_SYSTEM)
    /* kernel mode only */
    set_count |= PCR_SYS_TRACE;
  else
    /* user or kernel mode */
    set_count |= (PCR_SYS_TRACE | PCR_USER_TRACE);

  if(ioctl(hw_descr->fd, PERFMON_SETPCR, &set_count) < 0)
    return PCL_FAILURE;

  /* Clear and start counter */
  clr_pic();
  cpu_sync();

  /* read out counter values (used as base value) */
  val = get_pic();
  hw_descr->pic0_value = val & 0xffffffff;
  hw_descr->pic1_value = val >> 32;

  /* get actual clock value (to measure cycles) */
  hw_descr->clock_value = get_tick();

#else  /* defined(PERFMON) */

  /* Extract selected events */
  if(IS_USED(used_mask, 0))
    sel0 = commands[0];
  if(IS_USED(used_mask, 1))
    sel1 = commands[1];

  memset(&cpc_cmd, 0, sizeof(cpc_cmd));
  cpc_cmd.ce_cpuver = hw_descr->cpuver;
  cpc_cmd.ce_pcr =                        /* non-pivileged access possible */
                     (count_mode & PCL_MODE_SYSTEM) << 1    /* system trace */
                   | (count_mode & PCL_MODE_USER)   << 2    /* user trace */
                   | sel0 << 4           /* event selector 0 */
                   | sel1 << 11;         /* event selector 1 */

  if(cpc_bind_event(&cpc_cmd, 0) == -1)
    return PCL_FAILURE;

  if(cpc_take_sample(&cpc_cmd) == -1)
    return PCL_FAILURE;

  /* read out counter values (used as base value) */
  hw_descr->pic0_value = cpc_cmd.ce_pic[0];
  hw_descr->pic1_value = cpc_cmd.ce_pic[1];

#if 0
  /* ???
     Using get_tick() delivers an illegal instruction (and a core dump).
     With bit 0 of the PCR set to 0 in the call to cpc_bind_event 
     that should not happen. But mysteriously after the call to cpc_bind_event
     this bit is set in the data structure.
     What does SUN do in cpc_bind_event???
   */
  /* get actual clock value (to measure cycles) */
  hw_descr->clock_value = get_tick();
#else
  hw_descr->clock_value = cpc_cmd.ce_hrt;
#endif
#endif  /* defined(PERFMON) */


  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Read hardware counters                                                     */

int
PCL_driver_read(
		PCL_DESCR_TYPE d, /* I: descriptor */
		int max_counter,
		PCL_BIT_MASK_TYPE counter_used_mask,
		PCL_CNT_TYPE *counter_values
		)

{
#if defined(PERFMON)
  unsigned long long clocks;
#endif
  PCL_BIT_MASK_TYPE used_mask;
  PCL_state_descr *descr = d;
  HW_state_descr *hw_descr;
#if defined(PERFMON)
  unsigned long long val;
#else
  cpc_event_t cpc_cmd;
#endif


  /* get state descriptor */
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];


#if defined(PERFMON)
  /* finish all outstanding operations */
  cpu_sync();

  /* get actual cycle count */
  clocks = get_tick();

  /* get PIC-values as a difference between the actual value and the value
     gotten when starting the counts. The lower 32 bits are counter0 results,
     the upper 32 bits are counter1 results.
  */
  val = get_pic();
  if(IS_USED(used_mask, 0))
    counter_values[0] = (val & 0xffffffff) - hw_descr->pic0_value;
  if(IS_USED(used_mask, 1))
    counter_values[1] =  (val >> 32) - hw_descr->pic1_value;

  if(IS_USED(used_mask, 2))
    /* special case: cycle count. */
    counter_values[2] = clocks - hw_descr->clock_value;

#else  /* defined(PERFMON) */

#if 0
  /* finish all outstanding operations */
  cpu_sync();

  /* get actual cycle count */
  clocks = get_tick();

  /* get PIC-values as a difference between the actual value and the value
     gotten when starting the counts. The lower 32 bits are counter0 results,
     the upper 32 bits are counter1 results.
  */
  val = get_pic();
  if(IS_USED(used_mask, 0))
    counter_values[0] = (val & 0xffffffff) - hw_descr->pic0_value;
  if(IS_USED(used_mask, 1))
    counter_values[1] =  (val >> 32) - hw_descr->pic1_value;

  if(IS_USED(used_mask, 2))
    /* special case: cycle count. */
    counter_values[2] = clocks - hw_descr->clock_value;

#else
  /* ??? The call to cpc_take_sample has a HUGE overhead. It would be better to
     read out directly the performance registers but that is not possible
     due to the above described problem a la get_tick().
  */
  if(cpc_take_sample(&cpc_cmd) == -1)
    return PCL_FAILURE;

  if(IS_USED(used_mask, 0))
    counter_values[0] = cpc_cmd.ce_pic[0] - hw_descr->pic0_value;
  if(IS_USED(used_mask, 1))
    counter_values[1] =  cpc_cmd.ce_pic[1] - hw_descr->pic1_value;

  if(IS_USED(used_mask, 2))
    /* special case: cycle count. */
    /* hrt in ns */
    counter_values[2] = (cpc_cmd.ce_hrt - hw_descr->clock_value) * 1000 / hw_descr->mhz_rate;

#endif
#endif  /* defined(PERFMON) */


  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */


int
PCL_driver_stop(
		 PCL_DESCR_TYPE d /* I: descriptor */
		)

{
#if !defined(PERFMON)
  /* release performance counter context */
  if(cpc_rele() != 0)
    return PCL_FAILURE;
#endif

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* close                                                                      */

int
PCL_driver_close(
		 PCL_DESCR_TYPE d /* I: descriptor */
		 )

{
  /* ??? what happens here? */

  /* everything ok */
  return PCL_SUCCESS;
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
